/*
** File: flash.c  
** Project: ADB3 core driver
** Purpose: Implements functions for identifying, reading, writing and erasing the
**          Flash devices on a card.
**
** (C) Copyright Alpha Data 2009-2013
*/

#include "cfi.h"
#include "device.h"
#include "flash.h"
#include "flash_cfi.h"
#include "flash_legacy.h"

static CoreFlashStatus
eraseBlock(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int bankIndex,
  uint64_t blockAddress)
{
  dfDebugPrint(3, ("eraseBlock: entered, pDevCtx=%p bankIndex=%lu blockAddress=0x%08lx_%08lx\n",
    (void*)pDevCtx, (unsigned long)bankIndex, dfSplitUint64(blockAddress)));

  if (pDevCtx->info.flash[bankIndex].bLegacyFlash) {
    return flashLegacyEraseBlock(pDevCtx, bankIndex, blockAddress);
  } else {
    CfiInformation* pCfiInfo = pDevCtx->info.flash[bankIndex].detail.pCfi;

    switch (pCfiInfo->validation.preferredAlgorithm) {
    case CFI_ALG_INTELSHARP_EXT:
      return flashCfiEraseBlock0001(pDevCtx, bankIndex, blockAddress);

    case CFI_ALG_INTEL_PERF:
      return flashCfiEraseBlock0200(pDevCtx, bankIndex, blockAddress);

    default:
      /* Don't understand this command set */
      return CoreFlashGeneralFailure;
    }
  }
}

static CoreFlashStatus
readChunk(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int bankIndex,
  uint64_t start,
  uint64_t length,
  void* pBuffer)
{
  uint8_t* p;
  uint32_t val32;
  uint8_t align;
  uint64_t startAligned;

  dfDebugPrint(5, ("readChunk: entered, bankIndex=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx pBuffer=%p\n",
    (unsigned long)bankIndex, dfSplitUint64(start), dfSplitUint64(length), pBuffer));

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  if (length == 0) {
    return CoreFlashSuccess;
  }

  p = (uint8_t*)pBuffer;
  switch (pDevCtx->info.bootstrap.flash[bankIndex].width) {
  case 1U:
    /* Flash device has 8-bit data bus */
    startAligned = start;
    while (length) {
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)val32;
      length--;
      startAligned++;
    }
    break;

  case 2U:
    /* Flash device has 16-bit data bus*/
    startAligned = start & ~0x1ULL;
    align = (uint8_t)(start & 0x1U);
    /* Ensure flash address aligned to a 16-bit word */
    if (align) {
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 8U);
      length--;
      startAligned += 2;
    }
    /* Do bulk of 16-bit words */
    while (length >= 2) {
      /* Align 'start' to a 16-bit word */
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      length -= 2;
      startAligned += 2;
    }
    /* Finish last byte */
    if (length & 0x1U) {
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
    }
    break;

  case 4U:
    /* Flash device has 32-bit data bus */
    startAligned = start & ~0x3ULL;
    align = (uint8_t)(start & 0x3U);
    /* Ensure flash address aligned to a 32-bit word */
    switch (align) {
    case 0U:
      break;

    case 1U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 3;
      startAligned += 4;
      break;

    case 2U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 2;
      startAligned += 4;
      break;

    case 3U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 1;
      startAligned += 4;
      break;
    }
    /* Do bulk of 32-bit words */
    while (length >= 4) {
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      *p++ = (uint8_t)(val32 >> 24U);
      length -= 4;
      startAligned += 4;
    }
    /* Finish last few bytes */
    switch ((unsigned int)length & 0x3U) {
    case 0U:
      break;

    case 1U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      break;

    case 2U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      break;

    case 3U:
      pDevCtx->methods.pFlash(pDevCtx, bankIndex, FALSE, startAligned, &val32);
      *p++ = (uint8_t)(val32 >> 0U);
      *p++ = (uint8_t)(val32 >> 8U);
      *p++ = (uint8_t)(val32 >> 16U);
      break;
    }
    break;

  default:
    return CoreFlashGeneralFailure;
  }

  return CoreFlashSuccess;
}

static CoreFlashStatus
writeBlock(
  Adb3CoreDeviceContext*  pDevCtx,
  unsigned int bankIndex,
  uint64_t blockAddress,
  uint64_t blockSize,
  const uint8_t* pData)
{
  dfDebugPrint(3, ("writeBlock: entered, bankIndex=%lu blockAddress=0x%08lx_%08lx blockSize=0x%08lx_%08lx pData=%p\n",
    (unsigned long)bankIndex, dfSplitUint64(blockAddress), dfSplitUint64(blockSize), pData));

  if (pDevCtx->info.flash[bankIndex].bLegacyFlash) {
    return flashLegacyWriteBlock(pDevCtx, bankIndex, blockAddress, blockSize, pData);
  } else {
    CfiInformation* pCfiInfo = pDevCtx->info.flash[bankIndex].detail.pCfi;

    switch (pCfiInfo->validation.preferredAlgorithm) {
    case CFI_ALG_INTELSHARP_EXT:
      return flashCfiWriteBlock0001(pDevCtx, bankIndex, blockAddress, blockSize, pData);

    case CFI_ALG_INTEL_PERF:
      return flashCfiWriteBlock0200(pDevCtx, bankIndex, blockAddress, blockSize, pData);

    default:
      /* Don't understand this command set */
      return CoreFlashGeneralFailure;
    }
  }
}

/* Returns TRUE if the specified block is currently cached */
static boolean_t
bBlockIsInCache(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t blockAddress)
{
  return (pDevCtx->flashControl[bankIndex].cache.state != FlashCacheInvalid && pDevCtx->flashControl[bankIndex].cache.address == blockAddress) ? TRUE : FALSE;
}

/* If there is a dirty block cached, write it to the Flash device and transition to Valid */
static CoreFlashStatus
syncBlockInCache(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex)
{
  CoreFlashStatus status = CoreFlashSuccess;

  if (pDevCtx->flashControl[bankIndex].cache.state == FlashCacheDirty) {
    /* Erase the block in the Flash device so that it can be writtem with the data in the cache */
    status = eraseBlock(pDevCtx, bankIndex, pDevCtx->flashControl[bankIndex].cache.address);
    if (CoreFlashSuccess == status) {
      /* Write the cached block back to the Flash device */
      status = writeBlock(pDevCtx, bankIndex, pDevCtx->flashControl[bankIndex].cache.address, pDevCtx->flashControl[bankIndex].cache.length, pDevCtx->flashControl[bankIndex].cache.pBuffer);
      if (CoreFlashSuccess == status) {
        /* Transition to Valid state */
        pDevCtx->flashControl[bankIndex].cache.state = FlashCacheValid;
      } else {
        dfDebugPrint(0, ("*** syncBlockInCache: failed to write Flash block to hardware, bankIndex=%lu blockAddress=0x%08lx_%08lx blockSize=0x%08lx_%08lx pData=%p\n",
          (unsigned long)bankIndex,
          dfSplitUint64(pDevCtx->flashControl[bankIndex].cache.address),
          dfSplitUint64(pDevCtx->flashControl[bankIndex].cache.length),
          pDevCtx->flashControl[bankIndex].cache.pBuffer));
      }
    } else {
      dfDebugPrint(0, ("*** syncBlockInCache: failed to erase Flash block in hardware, bankIndex=%lu blockAddress=0x%08lx_%08lx\n",
        (unsigned long)bankIndex, dfSplitUint64(pDevCtx->flashControl[bankIndex].cache.address)));
    }
  }
  return status;
}

/* Ensure a block is resident in the cache */
static CoreFlashStatus
loadBlockIntoCache(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t blockStart,
  uint64_t blockLength)
{
  CoreFlashStatus status = CoreFlashSuccess;

  if (!bBlockIsInCache(pDevCtx, bankIndex, blockStart)) { 
    /* If there is a dirty block currently cached, sync it with the Flash device */
    status = syncBlockInCache(pDevCtx, bankIndex);
    if (CoreFlashSuccess != status) {
      return status;
    }
    /* Now read in the new cache block */
    status = readChunk(pDevCtx, bankIndex, blockStart, blockLength, pDevCtx->flashControl[bankIndex].cache.pBuffer);
    if (CoreFlashSuccess == status) {
      /* Transition to Valid state if successful */
      pDevCtx->flashControl[bankIndex].cache.address = blockStart;
      pDevCtx->flashControl[bankIndex].cache.length = blockLength;
      pDevCtx->flashControl[bankIndex].cache.state = FlashCacheValid;
    } else {
      /* An error occurred; ensure the cache is in Invalid state */
      pDevCtx->flashControl[bankIndex].cache.state = FlashCacheInvalid;
    }
  }
  return status;
}

/*
** -----------------------------------------------------------------
** Exported routines
** -----------------------------------------------------------------
*/

CoreFlashStatus
flashIdentifyBlock(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t location,
  uint64_t* pBlockStart,
  uint64_t* pBlockSize)
{
  Adb3CoreFlashInfo* pFlashInfo;

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  pFlashInfo = &pDevCtx->info.flash[bankIndex];
  if (!pFlashInfo->bPresent) {
    return CoreFlashNotPresent;
  }
  if (location > pFlashInfo->totalSize) {
    /* Location out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  if (pFlashInfo->bLegacyFlash) {
    LegacyFlashInformation* pLegacyInfo = pFlashInfo->detail.pLegacy;
    unsigned int i, n = pLegacyInfo->numSpecialBlock;
    uint32_t limit;

    /* Check if location is in a special/boot block */
    for (i = 0; i < n; i++) {
      limit = pLegacyInfo->specialBlock[i].address + (pLegacyInfo->specialBlock[i].size - 1);
      if (location >= pLegacyInfo->specialBlock[i].address && location <= limit) {
        /* Location is in a special/boot block */
        *pBlockStart = pLegacyInfo->specialBlock[i].address;
        *pBlockSize = pLegacyInfo->specialBlock[i].size;
        return CoreFlashSuccess;
      }
    }

    /* Location must be in a normal block */
    *pBlockStart = location & ~(uint64_t)(pLegacyInfo->blockSize - 1);
    *pBlockSize = pLegacyInfo->blockSize;
    return CoreFlashSuccess;
  } else {
    CfiInformation* pCfiInfo = pFlashInfo->detail.pCfi;
    uint64_t startAddr, endAddr;
    uint64_t blockIndex;

    while (NULL != pCfiInfo) {
      uint8_t i;

      if (location - pCfiInfo->chipAddress >= pCfiInfo->geometry.fixed.size) {
        /* The location is not in this die / device; advance to the next die */
        pCfiInfo = pCfiInfo->validation.pLink;
      } else {
        if (!pCfiInfo->validation.bBlockEraseSupported) {
          /* Only full-chip erase is supported; treat entire chip as a single block */
          *pBlockStart = pCfiInfo->chipAddress;
          *pBlockSize = pCfiInfo->geometry.fixed.size;
          return CoreFlashSuccess;
        }

        if (pCfiInfo->geometry.fixed.numEraseRegion == 0) {
          /* Should never get here, as zero erase regions => block-erase not supported */
          return CoreFlashGeneralFailure;
        }

        for (i = 0; i < pCfiInfo->geometry.fixed.numEraseRegion; i++) {
          startAddr = pCfiInfo->geometry.pEraseRegion[i].startAddress;
          endAddr = pCfiInfo->geometry.pEraseRegion[i].startAddress + pCfiInfo->geometry.pEraseRegion[i].blockSize * pCfiInfo->geometry.pEraseRegion[i].numBlock;
          if (location >= startAddr && location < endAddr) {
            blockIndex = dfDivide64By64(location - startAddr, pCfiInfo->geometry.pEraseRegion[i].blockSize);
            *pBlockStart = startAddr + blockIndex * pCfiInfo->geometry.pEraseRegion[i].blockSize;
            *pBlockSize = pCfiInfo->geometry.pEraseRegion[i].blockSize;
            return CoreFlashSuccess;
          }
        }

        /* If we get here, there must be an error in the data structure */
        return CoreFlashGeneralFailure;
      }
    }
  }

  /* The requested location is outside the bounds of the Flash device */
  return CoreFlashInvalidRegion;
}

CoreFlashStatus
flashErase(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t start,
  uint64_t length)
{
  CoreFlashStatus status = CoreFlashSuccess;
  Adb3CoreFlashInfo* pFlashInfo;
  Adb3CoreFlashControl* pFlashControl;
  uint64_t blockStart, blockSize, blockOffset;
  uint64_t i, chunk;
  uint8_t* pBuffer;

  dfDebugPrint(5, ("flashErase: bankIndex=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx\n",
    (unsigned long)bankIndex, dfSplitUint64(start), dfSplitUint64(length)));

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  pFlashInfo = &pDevCtx->info.flash[bankIndex];
  pFlashControl = &pDevCtx->flashControl[bankIndex];
  if (!pFlashInfo->bPresent) {
    return CoreFlashNotPresent;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pFlashInfo->totalSize || (start + length) > pFlashInfo->totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }
  
  pBuffer = pFlashControl->cache.pBuffer;

  while (length) {
    /* Find the address and size of the current block */
    status = flashIdentifyBlock(pDevCtx, bankIndex, start, &blockStart, &blockSize);
    if (CoreFlashSuccess != status) {
      return status;
    }
    blockOffset = start - blockStart;
    chunk = blockSize - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    if (chunk != blockSize) { /* Erasing less than a whole block - need to read the block & merge data */
      /* Ensure the block is resident in the cache */
      status = loadBlockIntoCache(pDevCtx, bankIndex, blockStart, blockSize);
      if (status != CoreFlashSuccess) {
        goto done;
      }
      /* Merge in 0xFF bytes for portion of the block to be erased */
      for (i = 0; i < chunk; i++) {
        pBuffer[blockOffset + i] = 0xFFU;
      }
      /* Transition to Dirty state */
      pFlashControl->cache.state = FlashCacheDirty;
    } else { /* Erasing an entire block */
      if (bBlockIsInCache(pDevCtx, bankIndex, blockStart)) {
        /* Block is currently cached, but we're erasing the whole block, so erase the block and transition to the Invalid state */
        status = eraseBlock(pDevCtx, bankIndex, blockStart);
        if (status != CoreFlashSuccess) {
          /* An error occurred; ensure the cache is in Invalid state */
          pFlashControl->cache.state = FlashCacheInvalid;
          goto done;
        }
        /* OK */
        pFlashControl->cache.state = FlashCacheInvalid;
      } else {
        /* Erasing a whole block that isn't cached; no need to change the cache state */
        status = eraseBlock(pDevCtx, bankIndex, blockStart);
        if (status != CoreFlashSuccess) {
          goto done;
        }
      }
    }
    start += chunk;
    length -= chunk;
  }

done:
  return status;
}

CoreFlashStatus
flashRead(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t start,
  size_t length,
  void* pBuffer)
{
  CoreFlashStatus status;
  Adb3CoreFlashInfo* pFlashInfo;
  Adb3CoreFlashControl* pFlashControl;
  uint64_t blockStart, blockSize, blockOffset;
  uint64_t chunk;
  uint8_t* p = (uint8_t*)pBuffer;

  dfDebugPrint(5, ("flashRead: bankIndex=%lu start=0x%lx length=%p pBuffer=%p\n",
    (unsigned long)bankIndex, (unsigned long)start, (void*)length, pBuffer));

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  pFlashInfo = &pDevCtx->info.flash[bankIndex];
  pFlashControl = &pDevCtx->flashControl[bankIndex];
  if (!pFlashInfo->bPresent) {
    return CoreFlashNotPresent;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pFlashInfo->totalSize || (start + length) > pFlashInfo->totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  while (length) {
    status = flashIdentifyBlock(pDevCtx, bankIndex, start, &blockStart, &blockSize);
    if (CoreFlashSuccess != status) {
      return status;
    }
    blockOffset = start - blockStart;
    chunk = blockSize - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    if (bBlockIsInCache(pDevCtx, bankIndex, blockStart)) {
      /* Get data from cache */
      dfCopyKK(p, pFlashControl->cache.pBuffer + (size_t)blockOffset, (size_t)chunk /* cast is safe as chunk <= length */);
    } else {
      /* Get data from Flash device */
      status = readChunk(pDevCtx, bankIndex, start, chunk, p);
      if (CoreFlashSuccess != status) {
        return status;
      }
    }
    start += chunk;
    length -= (size_t)chunk; /* cast is safe as chunk <= length */
    p += (size_t)chunk; /* cast is safe as chunk <= length */
  }

  return CoreFlashSuccess;
}

CoreFlashStatus
flashWrite(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex,
  uint64_t start,
  size_t length,
  const void* pData)
{
  CoreFlashStatus status = CoreFlashSuccess;
  Adb3CoreFlashInfo* pFlashInfo;
  Adb3CoreFlashControl* pFlashControl;
  uint64_t blockStart, blockSize, blockOffset;
  uint64_t i, chunk;
  uint8_t* p = (uint8_t*)pData;
  uint8_t* pBuffer;

  dfDebugPrint(5, ("flashWrite: bankIndex=%lu start=0x%08lx_%08lx length=0x%08lx_%08lx pData=%p\n",
    (unsigned long)bankIndex, dfSplitUint64(start), dfSplitUint64((uint64_t)length), pData));

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  pFlashInfo = &pDevCtx->info.flash[bankIndex];
  pFlashControl = &pDevCtx->flashControl[bankIndex];
  if (!pFlashInfo->bPresent) {
    return CoreFlashNotPresent;
  }
  if (start + length < start) {
    /* Region wraps to 0 */
    return CoreFlashInvalidRegion;
  }
  if (start > pFlashInfo->totalSize || (start + length) > pFlashInfo->totalSize) {
    /* Region out of bounds of Flash chip */
    return CoreFlashInvalidRegion;
  }

  pBuffer = pFlashControl->cache.pBuffer;

  while (length) {
    /* Find the address and size of the current block */
    status = flashIdentifyBlock(pDevCtx, bankIndex, start, &blockStart, &blockSize);
    if (CoreFlashSuccess != status) {
      return status;
    }
    blockOffset = start - blockStart;
    chunk = blockSize - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    if (chunk != blockSize) { /* Less than entire block is to be written - need to read the block & merge data */
      /* Ensure the block is resident in the cache */
      status = loadBlockIntoCache(pDevCtx, bankIndex, blockStart, blockSize);
      if (status != CoreFlashSuccess) {
        goto done;
      }
      /* Merge in new data */
      for (i = 0; i < chunk; i++) {
        pBuffer[blockOffset + i] = *p++;
      }
      /* Transition to Dirty state */
      pFlashControl->cache.state = FlashCacheDirty;
    } else { /* The entire block is being written */
      if (bBlockIsInCache(pDevCtx, bankIndex, blockStart)) {
        /* The currently cached block is being written in its entirety */
        for (i = 0; i < chunk; i++) {
          pBuffer[blockOffset + i] = *p++;
        }
        status = eraseBlock(pDevCtx, bankIndex, blockStart);
        if (status != CoreFlashSuccess) {
          /* An error occurred, so ensure the cache is marked Invalid */
          pFlashControl->cache.state = FlashCacheInvalid;
          goto done;
        }
        status = writeBlock(pDevCtx, bankIndex, blockStart, blockSize, pBuffer);
        if (status != CoreFlashSuccess) {
          /* An error occurred, so ensure the cache is marked Invalid */
          pFlashControl->cache.state = FlashCacheInvalid;
          goto done;
        }
        /* Transition to Valid state */
        pFlashControl->cache.state = FlashCacheValid;
      } else {
        status = eraseBlock(pDevCtx, bankIndex, blockStart);
        if (status != CoreFlashSuccess) {
          goto done;
        }
        /* An entire block that is not in the cache is being written; no need to change cache state */
        status = writeBlock(pDevCtx, bankIndex, blockStart, blockSize, p);
        if (status != CoreFlashSuccess) {
          goto done;
        }
        p += blockSize;
      }
    }
    start += chunk;
    length -= (size_t)chunk; /* Cast is safe since we forced chunk <= length above */
  }

done:
  return status;
}

CoreFlashStatus
flashSync(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int bankIndex)
{
  Adb3CoreFlashInfo* pFlashInfo;

  dfDebugPrint(5, ("flashSync: bankIndex=%u\n", bankIndex));

  if (bankIndex >= pDevCtx->info.bootstrap.numFlashBank) {
    return CoreFlashInvalidIndex;
  }
  pFlashInfo = &pDevCtx->info.flash[bankIndex];
  if (!pFlashInfo->bPresent) {
    return CoreFlashNotPresent;
  }

  return syncBlockInCache(pDevCtx, bankIndex);
}
