/*
** File: pio.c  
** Project: ADMXRC2 module driver
** Purpose: OS-independent IOCTL handlers for programmed I/O.
**
** (C) Copyright Alpha Data 2013
*/

#include <df.h>

#include <core/coreif.h>
#include "device.h"
#include "pio.h"
#include <admxrc2/types.h>

/* Length threshold at which we install a cancel callback for I/O requests */
#define CANCEL_THRESHOLD (0x1000U)

typedef struct _Context {
  boolean_t bCancel;
  DfIoStatus reason;
  DfEvent event;
} Context;

#define MAX_CHUNK 256U
#define MAX_CHUNK8 (MAX_CHUNK)
#define MAX_CHUNK16 (MAX_CHUNK >> 1)
#define MAX_CHUNK32 (MAX_CHUNK >> 2)
#define MAX_CHUNK64 (MAX_CHUNK >> 3)

static ADMXRC2_STATUS
findWindow(
	Admxrc2DeviceContext* pDevCtx,
  uint32_t width,
  uint32_t flags,
  uint32_t local,
  size_t length,
  uint8_t** ppKernelAddr,
  DfMemoryHandle* phMemory)
{
  unsigned int i;
  WindowAccessInfo* pAccessInfo;
  uint64_t localLimit;
  uint32_t offset;
  size_t accessLimit;
  uint8_t* pKernelAddr;

  if (flags & ADMXRC2_IOFIXED) {
    accessLimit = local + width;
  } else {
    accessLimit = local + length;
  }

  if (flags & ADMXRC2_IOADAPTER) {
    pKernelAddr = (uint8_t*)pDevCtx->window.adapter.pKernelBase;
    if (NULL == pKernelAddr) {
      /* Adapter BAR not found, shouldn't happen, but it did... */
      /* TO DO - decide if return ADMXRC_INVALID_REGION or keep old behavior of ADMXRC_INVALID_PARAMETER */
      return ADMXRC2_INVALID_REGION;
    }
    if (accessLimit > pDevCtx->window.adapter.size) {
      /* Access length extends past end of adapter register BAR */
      /* TO DO - decide if return ADMXRC2_REGION_TOO_LARGE or keep old behavior of ADMXRC_INVALID_PARAMETER */
      return ADMXRC2_REGION_TOO_LARGE;
    }
    /* Access length for adapter register BAR was validated OK */
    *ppKernelAddr = pKernelAddr + local;
    *phMemory = pDevCtx->window.adapter.hMemory;
    return ADMXRC2_SUCCESS;
  } else {
    for (i = 0; i < pDevCtx->info.numWindow; i++) {
      pAccessInfo = &pDevCtx->window.access[i];
      if (0 == pAccessInfo->size) {
        /* Must be adapter register BAR with no local bus presence - ignore */
        continue;
      }
      localLimit = pAccessInfo->localBase + pAccessInfo->size;
      if (pAccessInfo->localBase <= local && local < localLimit) {
        offset = local - (uint32_t)pAccessInfo->localBase;
        pKernelAddr = (uint8_t*)pAccessInfo->pKernelBase + offset;
        if (accessLimit > localLimit) {
          /* Local window containing 'local' was found, but access length extends past end of window */
          /* TO DO - decide if return ADMXRC2_REGION_TOO_LARGE or keep old behavior of ADMXRC_INVALID_PARAMETER */
          return ADMXRC2_REGION_TOO_LARGE;
        }
        /* Found a local window containing 'local' and access length was validated OK */
        *ppKernelAddr = pKernelAddr;
        *phMemory = pAccessInfo->hMemory;
        return ADMXRC2_SUCCESS;
      }
    }
    /* Local window containing 'local' was not found */
    /* TO DO - decide if return ADMXRC_INVALID_REGION or keep old behavior of ADMXRC_INVALID_PARAMETER */
    return ADMXRC2_INVALID_REGION;
  }
}

static void
cancelCallback(
  DfClientRequest* pReq,
  DfIoStatus reason)
{
  Context* pContext = (Context*)dfRequestGetContext(pReq);

  pContext->reason = reason;
  pContext->bCancel = TRUE;
  dfEventSignal(&pContext->event);
}

DfIoStatus
ioctlRead(
	Admxrc2DeviceContext* pDevCtx,
	Admxrc2ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC2_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = ADMXRC2_IOFIXED | ADMXRC2_IOADAPTER;
#endif
  IOCTLS_ADMXRC2_READ* pIoctl = (IOCTLS_ADMXRC2_READ*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  Context context;
  uint32_t widthOrder, local, flags;
  boolean_t bFixedLocalAddress;
  size_t length, limit, remaining, remainingWords;
  unsigned int width, widthMask, offset, preamble, nIndex, chunk, chunkWords;
  uint8_t* pUserBuf;
  uint8_t* pKernelAddr;
  DfMemoryHandle hMemory;
  union _Value {
    uint8_t preamble[8];
    uint8_t val8[MAX_CHUNK8];
    uint16_t val16[MAX_CHUNK16];
    uint32_t val32[MAX_CHUNK32];
    uint64_t val64[MAX_CHUNK64];
    uint8_t residue[8];
  } val;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!admxrc2CheckAccessRights(pClCtx)) {
    return DfIoStatusError(ADMXRC2_ACCESS_DENIED);
  }

  widthOrder = pIoctl->in.width;
  if (widthOrder > ADMXRC2_IOWIDTH_64) {
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }
#if !DF_64BIT
  if (widthOrder == ADMXRC2_IOWIDTH_64) {
    /* If not running on a 64-bit architecture and 64-bit width is requested, fall back to 32-bit width */
    widthOrder = ADMXRC2_IOWIDTH_32;
  }
#endif
  width = 1U << widthOrder;
  flags = pIoctl->in.flags;
  local = pIoctl->in.local;
  pUserBuf = pIoctl->in.pBuffer;
  length = pIoctl->in.length;
#if defined(ADMXRC2_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }
#endif

  widthMask = width - 1U;
  bFixedLocalAddress = (flags & ADMXRC2_IOFIXED) ? TRUE : FALSE;
  offset = (unsigned int)local & widthMask;
  limit = length + (size_t)local;

  dfDebugPrint(6, ("ioctlRead: widthOrder=%lu flags=0x%lx local=0x%lx pUserBuf=%p length=%p\n",
    (unsigned long)widthOrder, (unsigned long)flags, (unsigned long)local, pUserBuf, (uintptr_t)length));

  status = findWindow(pDevCtx, width, flags, local, length, &pKernelAddr, &hMemory);
  if (ADMXRC2_SUCCESS != status) {
    return status;
  }

  remaining = length;

  if (bFixedLocalAddress && (offset || (limit & widthMask))) {
    /* If using fixed local bus address, local bus address must be aligned */
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }

  context.bCancel = FALSE;
  if (length >= CANCEL_THRESHOLD) {
    dfRequestSetContext(pReq, &context);
    dfEventInit(&context.event);
    if (!dfRequestSetCallback(pReq, cancelCallback)) {
      return DfIoStatusError(ADMXRC2_CANCELLED); 
    }
  }

  DF_USER_SPACE_BEGIN(TAG1)

  dfUserBufferValidate(pUserBuf, length, TRUE, TAG1);

  /* If start local address is unaligned, do first few bytes as memcpy to get aligned */
  if (offset) {
    preamble = width - offset;
    chunk = (remaining > preamble) ? preamble : (unsigned int)remaining;
    for (nIndex = 0; nIndex < chunk; nIndex++) {
      val.preamble[nIndex] = dfPciMemRead8(hMemory, pKernelAddr);
      pKernelAddr++;
    }
    dfCopyKU(pUserBuf, val.preamble, chunk, TAG1);
    remaining -= chunk;
    pUserBuf += chunk;
  }

  /* At this point, pKernelAddr is aligned to width */
  switch (widthOrder) {
#if DF_64BIT
    case ADMXRC2_IOWIDTH_64:
      remainingWords = remaining >> 3;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK64) ? MAX_CHUNK64 : (unsigned int)remainingWords;
        chunk = chunkWords << 3;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          val.val64[nIndex] = dfPciMemRead64(hMemory, (uint64_t*)pKernelAddr);
          if (!bFixedLocalAddress) {
            pKernelAddr += 8;
          }
        }
        dfCopyKU(pUserBuf, val.val64, chunk, TAG1);
        pUserBuf += chunk;
        remainingWords -= chunkWords;
      }
      remaining &= 0x7;
      break;
#endif

    case ADMXRC2_IOWIDTH_32:
      remainingWords = remaining >> 2;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK32) ? MAX_CHUNK32 : (unsigned int)remainingWords;
        chunk = chunkWords << 2;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          val.val32[nIndex] = dfPciMemRead32(hMemory, (uint32_t*)pKernelAddr);
          if (!bFixedLocalAddress) {
            pKernelAddr += 4;
          }
        }
        dfCopyKU(pUserBuf, val.val32, chunk, TAG1);
        pUserBuf += chunk;
        remainingWords -= chunkWords;
      }
      remaining &= 0x3;
      break;

    case ADMXRC2_IOWIDTH_16:
      remainingWords = remaining >> 1;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK16) ? MAX_CHUNK16 : (unsigned int)remainingWords;
        chunk = chunkWords << 1;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          val.val16[nIndex] = dfPciMemRead16(hMemory, (uint16_t*)pKernelAddr);
          if (!bFixedLocalAddress) {
            pKernelAddr += 2;
          }
        }
        dfCopyKU(pUserBuf, val.val16, chunk, TAG1);
        pUserBuf += chunk;
        remainingWords -= chunkWords;
      }
      remaining &= 0x1;
      break;

    case ADMXRC2_IOWIDTH_8:
    default:
      while (remaining && !context.bCancel) {
        chunk = (remaining > MAX_CHUNK8) ? MAX_CHUNK8 : (unsigned int)remaining;
        for (nIndex = 0; nIndex < chunk; nIndex++) {
          val.val8[nIndex] = dfPciMemRead8(hMemory, pKernelAddr);
          if (!bFixedLocalAddress) {
            pKernelAddr++;
          }
        }
        dfCopyKU(pUserBuf, val.val8, chunk, TAG1);
        pUserBuf += chunk;
        remaining -= chunk;
      }
      break;
  }

  /* If end local address is unaligned, do last few bytes as memcpy */
  if (remaining && !context.bCancel) {
    for (nIndex = 0; nIndex < remaining; nIndex++) {
      val.residue[nIndex] = dfPciMemRead8(hMemory, pKernelAddr);
      pKernelAddr++;
    }
    dfCopyKU(pUserBuf, val.residue, remaining, TAG1);
  }

  DF_USER_SPACE_EXCEPT(TAG1)

  status = DfIoStatusError(ADMXRC2_INVALID_PARAMETER); /* TODO - could return better status code here */

  DF_USER_SPACE_END(TAG1)

  if (length >= CANCEL_THRESHOLD) {
    if (!dfRequestClearCallback(pReq)) {
      dfEventWait(&context.event);
      status = context.reason;
    }
    dfEventUninit(&context.event);
  }

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadThunk(
	Admxrc2DeviceContext* pDevCtx,
	Admxrc2ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC2_READ* pIoctl32 = (IOCTLS32_ADMXRC2_READ*)pBuffer;
  IOCTLS_ADMXRC2_READ ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.width = pIoctl32->in.width;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.local = pIoctl32->in.local;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  ioctl.in.length = (size_t)pIoctl32->in.length;
  return ioctlRead(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWrite(
	Admxrc2DeviceContext* pDevCtx,
	Admxrc2ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC2_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = ADMXRC2_IOFIXED | ADMXRC2_IOADAPTER;
#endif
  IOCTLS_ADMXRC2_WRITE* pIoctl = (IOCTLS_ADMXRC2_WRITE*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  Context context;
  uint32_t widthOrder, local, flags;
  size_t length, limit, remaining, remainingWords;
  boolean_t bFixedLocalAddress;
  unsigned int width, widthMask, offset, preamble, nIndex, chunk, chunkWords;
  const uint8_t* pUserBuf;
  uint8_t* pKernelAddr;
  DfMemoryHandle hMemory;
  union _Value {
    uint8_t preamble[8];
    uint8_t val8[MAX_CHUNK8];
    uint16_t val16[MAX_CHUNK16];
    uint32_t val32[MAX_CHUNK32];
    uint64_t val64[MAX_CHUNK64];
    uint8_t residue[8];
  } val;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!admxrc2CheckAccessRights(pClCtx)) {
    return DfIoStatusError(ADMXRC2_ACCESS_DENIED);
  }

  widthOrder = pIoctl->in.width;
  if (widthOrder > ADMXRC2_IOWIDTH_64) {
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }
#if !DF_64BIT
  if (widthOrder == ADMXRC2_IOWIDTH_64) {
    /* If not running on a 64-bit architecture and 64-bit width is requested, fall back to 32-bit width */
    widthOrder = ADMXRC2_IOWIDTH_32;
  }
#endif
  width = 1U << widthOrder;
  flags = pIoctl->in.flags;
  local = pIoctl->in.local;
  pUserBuf = (const uint8_t*)pIoctl->in.pData;
  length = pIoctl->in.length;
#if defined(ADMXRC2_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }
#endif

  widthMask = width - 1U;
  bFixedLocalAddress = (flags & ADMXRC2_IOFIXED) ? TRUE : FALSE;
  offset = (unsigned int)local & widthMask;
  limit = length + (size_t)local;

  dfDebugPrint(6, ("ioctlWrite: widthOrder=%lu flags=0x%lx local=0x%lx pUserBuf=%p length=%p\n",
    (unsigned long)widthOrder, (unsigned long)flags, (unsigned long)local, pUserBuf, (uintptr_t)length));

  status = findWindow(pDevCtx, width, flags, local, length, &pKernelAddr, &hMemory);
  if (ADMXRC2_SUCCESS != status) {
    return status;
  }

  remaining = length;

  if (bFixedLocalAddress && (offset || (limit & widthMask))) {
    /* If using fixed local bus address, local bus address must be aligned */
    return DfIoStatusError(ADMXRC2_INVALID_PARAMETER);
  }

  context.bCancel = FALSE;
  if (length >= CANCEL_THRESHOLD) {
    dfRequestSetContext(pReq, &context);
    dfEventInit(&context.event);
    if (!dfRequestSetCallback(pReq, cancelCallback)) {
      return DfIoStatusError(ADMXRC2_CANCELLED); 
    }
  }

  DF_USER_SPACE_BEGIN(TAG1)

  dfUserBufferValidate(pUserBuf, length, FALSE, TAG1);

  /* If start local address is unaligned, do first few bytes individually to get aligned */
  if (offset) {
    preamble = width - offset;
    chunk = (remaining > preamble) ? preamble : (unsigned int)remaining;
    dfCopyUK(val.preamble, pUserBuf, chunk, TAG1);
    pUserBuf += chunk;
    remaining -= chunk;
    for (nIndex = 0; nIndex < chunk; nIndex++) {
      dfPciMemWrite8(hMemory, pKernelAddr, val.preamble[nIndex]);
      pKernelAddr++;
    }
  }

  /* At this point, pKernelAddr is aligned to width */
  switch (widthOrder) {
#if DF_64BIT
    case ADMXRC2_IOWIDTH_64:
      remainingWords = remaining >> 3;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK64) ? MAX_CHUNK64 : (unsigned int)remainingWords;
        chunk = chunkWords << 3;
        dfCopyUK(val.val64, pUserBuf, chunk, TAG1);
        pUserBuf += chunk;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          dfPciMemWrite64(hMemory, (uint64_t*)pKernelAddr, val.val64[nIndex]);
          if (!bFixedLocalAddress) {
            pKernelAddr += 8;
          }
        }
        remainingWords -= chunkWords;
      }
      remaining &= 0x7;
      break;
#endif

    case ADMXRC2_IOWIDTH_32:
      remainingWords = remaining >> 2;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK32) ? MAX_CHUNK32 : (unsigned int)remainingWords;
        chunk = chunkWords << 2;
        dfCopyUK(val.val32, pUserBuf, chunk, TAG1);
        pUserBuf += chunk;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          dfPciMemWrite32(hMemory, (uint32_t*)pKernelAddr, val.val32[nIndex]);
          if (!bFixedLocalAddress) {
            pKernelAddr += 4;
          }
        }
        remainingWords -= chunkWords;
      }
      remaining &= 0x3;
      break;

    case ADMXRC2_IOWIDTH_16:
      remainingWords = remaining >> 1;
      while (remainingWords && !context.bCancel) {
        chunkWords = (remainingWords > MAX_CHUNK16) ? MAX_CHUNK16 : (unsigned int)remainingWords;
        chunk = chunkWords << 1;
        dfCopyUK(val.val16, pUserBuf, chunk, TAG1);
        pUserBuf += chunk;
        for (nIndex = 0; nIndex < chunkWords; nIndex++) {
          dfPciMemWrite16(hMemory, (uint16_t*)pKernelAddr, val.val16[nIndex]);
          if (!bFixedLocalAddress) {
            pKernelAddr += 2;
          }
        }
        remainingWords -= chunkWords;
      }
      remaining &= 0x1;
      break;

    case ADMXRC2_IOWIDTH_8:
    default:
      while (remaining && !context.bCancel) {
        chunk = (remaining > MAX_CHUNK) ? MAX_CHUNK : (unsigned int)remaining;
        dfCopyUK(val.val8, pUserBuf, chunk, TAG1);
        pUserBuf += chunk;
        for (nIndex = 0; nIndex < chunk; nIndex++) {
          dfPciMemWrite8(hMemory, pKernelAddr, val.val8[nIndex]);
          if (!bFixedLocalAddress) {
            pKernelAddr++;
          }
        }
        remaining -= chunk;
      }
      break;
  }

  /* If end local address is unaligned, do last few bytes individually */
  if (remaining && !context.bCancel) {
    dfCopyUK(val.residue, pUserBuf, remaining, TAG1);
    for (nIndex = 0; nIndex < remaining; nIndex++) {
      dfPciMemWrite8(hMemory, pKernelAddr, val.residue[nIndex]);
      pKernelAddr++;
    }
  }

  DF_USER_SPACE_EXCEPT(TAG1)

  status = DfIoStatusError(ADMXRC2_INVALID_PARAMETER); /* TODO - could return better status code here */

  DF_USER_SPACE_END(TAG1)

  if (length >= CANCEL_THRESHOLD) {
    if (!dfRequestClearCallback(pReq)) {
      dfEventWait(&context.event);
      status = context.reason;
    }
    dfEventUninit(&context.event);
  }

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteThunk(
	Admxrc2DeviceContext* pDevCtx,
	Admxrc2ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC2_WRITE* pIoctl32 = (IOCTLS32_ADMXRC2_WRITE*)pBuffer;
  IOCTLS_ADMXRC2_WRITE ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.width = pIoctl32->in.width;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.local = pIoctl32->in.local;
  ioctl.in.pData = dfThunkConstPtr(pIoctl32->in.pData);
  ioctl.in.length = (size_t)pIoctl32->in.length;
  return ioctlWrite(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif